/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: parser_to_core.h,v $
*	$Revision: 1.6 $
*	$Date: 2006/12/28 13:17:37 $
*	$Author: zapreevis $
*	
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*	
*	Copyright (C) The University of Twente, 2004-2006.
*	Authors: Maneesh Khattri, Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is an intermediate interface between
*	the parser and the core model checkier.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#include "macro.h"
#include "foxglynn.h"
#include "bitset.h"
#include "sparse.h"
#include "label.h"
#include "runtime.h"
#include "transient.h"
#include "steady.h"
#include "prctl.h"
#include "help.h"

#ifndef PARSET_TO_CORE
#define PARSER_TO_CORE

/**
* This procedure sets the bitset result of the stateformula into the runtime.c
* using the set_result_bitset(pResultBitset).
* This method is alse used for printing the resulting probabilities if
* any and the set of states that satisfy the formula.
* NOTE: Printing is done only in case when the call of isPrintingOn()
* returns TRUE, get_result_probs() != NULL or get_result_bitset() != NULL
* @param pResultBitset: the satisfyability set for the stateformula.
*/
extern inline void setAndPrintResults( bitset * pResultBitset );

/**
* This method does printing of the error messages stored in the 
* help.h file.
*/
extern inline void printHelpMessage();
/**
* This method is used for printing state probability if any.
* In case get_result_probs_size() == 0, get_result_probs() == NULL
* or the index is out of bounds, the error messages are printed.
* @param index: the state index >= 1
*/
extern inline void printResultingStateProbability( const int index);
/**
* This method is used for printing state satisfyability result if any.
* I.e. if the state with the given index satisfyes the previously checked
* stateformula. In case get_result_bitset() == NULL or the index is out
* of bounds, the error messages are printed.
* @param index: the state index >= 1
*/
extern inline printResultingStateSatisfyability( const int index );

/**
* This method creates and returns the bitset filled with ones.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @return the bitset of the length get_labeller()->ns that
*         contains all states, i.e. every bit in the set is assigned to 1.
*/
extern inline bitset * createTrueBitset();

/**
* This method creates and returns the bitset filled with zero.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @return the bitset of the length get_labeller()->ns that
*         contains no states, i.e. every bit in the set is assigned to 0.
*/
extern inline bitset * createFalseBitset();

/**
* This method returns a set of states that satisfy atomic proposition 'label'.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @param label: the atomic proposition name.
* @return the set of states satisfying the atomic proposition 'label'.
*/
extern inline bitset * getStatesSetByLabel( const char * label );

/**
* This function is used for processing the formula
* connectives like AND, OR, IMPLIES. We have it separated
* because it is used in too many places here.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* NOTE: The input bitsets 'one' and 'two' are freed here.
* @param one: the first argument
* @param connective: the type of connective
* @param two: the second argument
* @result: the resulting bitset for 'one connective two'
*/
extern inline bitset * doConnective(bitset * one, int connective, bitset * two);

/**
* This method is used to compute the complement of a bitset.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @param pBitset: the input set
* @return the complement of the pBitset
*/
extern inline bitset * getComplementBitset(bitset * pBitset);

/**
* This function call model checking procedure for the steady state operator and then
* computes the set of states that satisfy the entire formula.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* @param pBitsetStateFormula: the state formula result
* @param probBound: the probability bound for the S operator
* @param comparator: the comparator id from the parser
* @return the set of states satysfying the S formula
*/
extern inline bitset * getSteadyStateOperatorResultBitset(bitset * pBitsetStateFormula, double probBound, int comparator);

/**
* This function computes the set of states that satisfy the P formula.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* @param result: the path formula result
* @param probBound: the probability bound for the P operator
* @param comparator: the comparator id from the parser
* @return the set of states satysfying the P formula
*/
extern inline bitset * getProbabilityOperatorResultBitset(double * result, double probBound, int comparator);

/**
* This function calls model-checking algorithm for the unbounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @return the array of probabilities
*/
extern inline double * getUnboundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset);

/**
* This function calls model-checking algorithm for the time-bounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @return the array of probabilities
*/
extern inline double * getTimeBoundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset, double t_bound_1, double t_bound_2);

/**
* This function calls model-checking algorithm for the time- and reward-bounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the array of probabilities
*/
extern inline double * getTimeAndRewardBoundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset, double t_bound_1, double t_bound_2, double r_bound_1, double r_bound_2);

/**
* This function calls model-checking algorithm for the next formula
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X phi' is checked
* @return the array of probabilities
*/
extern inline double * getNextProbability(bitset * pBitset);

/**
* This function calls model-checking algorithm for the time-bounded next formula
* NOTE: the pBitset is freed here
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @return the array of probabilities
*/
extern inline double * getTimeBoundedNextProbability(bitset * pBitset, double t_bound_1, double t_bound_2);

/**
* This function calls model-checking algorithm for the time- and reward-bounded next formula
* NOTE: the pBitset is freed here
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @return the array of probabilities
*/
extern inline double * getTimeAndRewardBoundedNextProbability(bitset * pBitset, double t_bound_1, double t_bound_2, double r_bound_1, double r_bound_2);
/**
* Calls model checking of the expected reward rate formula E.
* NOTE For the long run expected reward rate the 'epoch' parameter should be set to 0 (ZERO)
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
extern inline bitset * getExpectedRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2);

/**
* Calls model checking of the instantaneous reward formula C.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
extern inline bitset * getInstantaneousRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2);

/**
* Calls model checking of the expected accumulated reward Y.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
extern inline bitset * getExpectedAccumulatedRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2);

/**
* Start the timer.
* NOTE: The start of the timer should be done only in case we have an atomic
* proposition, 'tt' or 'ff' formula, becuase model checking starts from the leafs.
* WARNING: BE CAREFULL USING THIS MEHTOD. THERE SHOULD BE NO CONCURRENT USE OF
* THE TIMER FROM DIFFERENT ALGORITHMS AND SUCH.
* For example if we use this method for monitoring elapsed time of model checking and
* formula independent lumping, which do not overlap in time!
*/
extern inline void startTimer();

/**
* Stop the timer and return the elapsed time in milliseconds.
* NOTE: The stop of the timer should be done only in case the top level
* state formula procedure is called. I.e. stateformula NEWLINE.
* WARNING: BE CAREFULL USING THIS MEHTOD. THERE SHOULD BE NO CONCURRENT USE OF
* THE TIMER FROM DIFFERENT ALGORITHMS AND SUCH.
* For example if we use this method for monitoring elapsed time of model checking and
* formula independent lumping, which do not overlap in time!
* @return elapsed time in milliseconds
*/
extern inline long stopTimer();

/**
* This method should be used in case where cleaning of all previously
* gathered results is necessary.
*/
extern inline void cleanResults();

#endif
