/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: parser_to_core.c,v $
*	$Revision: 1.9 $
*	$Date: 2007/08/17 12:03:28 $
*	$Author: zapreevis $
*	
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*	
*	Copyright (C) The University of Twente, 2004-2006.
*	Authors: Maneesh Khattri, Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is an intermediate interface between
*	the parser and the core model checkier.
*/

#include "parser_to_core.h"
#include "y.tab.h"

//This variable stores the time when model checking has started
static struct timeval start_time = {0,0};

/**
* Start the timer.
* NOTE: Here the start of the timer should be done only in case we have an atomic
* proposition, 'tt' or 'ff' formula, becuase model checking starts from the leafs.
* WARNING: BE CAREFULL USING THIS MEHTOD. THERE SHOULD BE NO CONCURRENT USE OF
* THE TIMER FROM DIFFERENT ALGORITHMS AND SUCH.
* For example if we use this method for monitoring elapsed time of model checking and
* formula independent lumping, which do not overlap in time!
*/
inline void startTimer(){
	//If timer has not been started yet
	if( start_time.tv_sec == 0 && start_time.tv_usec == 0 ){
		gettimeofday(&start_time, NULL);
	}
}

/**
* Stop the timer and return the elapsed time in milliseconds.
* NOTE: Here the stop of the timer should be done only in case the top level
* state formula procedure is called. I.e. stateformula NEWLINE.
* WARNING: BE CAREFULL USING THIS MEHTOD. THERE SHOULD BE NO CONCURRENT USE OF
* THE TIMER FROM DIFFERENT ALGORITHMS AND SUCH.
* For example if we use this method for monitoring elapsed time of model checking and
* formula independent lumping, which do not overlap in time!
* @return elapsed time in milliseconds
*/
inline long stopTimer(){
	long elapsed_time = 0;
	struct timeval end_time;
	
	//If timer has been started
	if( start_time.tv_sec != 0 && start_time.tv_usec != 0 ){
		gettimeofday(&end_time, NULL);
		
		elapsed_time = ( end_time.tv_sec - start_time.tv_sec ) * 1000 + ( end_time.tv_usec - start_time.tv_usec ) / 1000;
		
		//Reset start_time to zero
		start_time.tv_sec = 0;
		start_time.tv_usec = 0;
	}
	return elapsed_time;
}

/**
* This procedure sets the bitset result of the stateformula into the runtime.c
* using the set_result_bitset(pResultBitset).
* This method is alse used for printing the resulting probabilities if
* any and the set of states that satisfy the formula.
* NOTE: Printing is done only in case when the call of isPrintingOn()
* returns TRUE, get_result_probs() != NULL or get_result_bitset() != NULL
* @param pResultBitset: the satisfyability set for the stateformula.
*/
inline void setAndPrintResults(bitset * pResultBitset){
	set_result_bitset( pResultBitset );
	const long elapsed_time = stopTimer();
	
	if( isPrintingOn() ){
		double *result = get_result_probs();
		int size = get_result_probs_size();
		if( result != NULL && size > 0){
			print_state_probs(size, result);
		}
		
		bitset *b = get_result_bitset();
		if( b != NULL ){
			print_states(b);
		}
	}
	
	printf("The Total Elapsed Model-Checking Time is %ld milli sec(s).\n", elapsed_time);
}

/**
* This procedure does printing of the error messages stored in the 
* help.h file.
*/
inline void printHelpMessage(){
	const unsigned int mode = isRunMode(ANY_LOGIC_MODE);
	switch(mode)
	{
		case CSL_MODE:
			printf("%s",HELP_CSL_MSG);
			break;
		case PCTL_MODE:
			printf("%s",HELP_PCTL_MSG);
			break;
		case PRCTL_MODE:
			printf("%s",HELP_PRCTL_MSG);
			break;
		case CSRL_MODE:
			printf("%s",HELP_CSRL_MSG);
			break;
		default :
			printf("%s",HELP_MSG);
	}
}

/**
* This method is used for printing state probability if any.
* In case get_result_probs_size() == 0, get_result_probs() == NULL
* or the index is out of bounds, the error messages are printed.
* @param index: the state index >= 1
*/
inline void printResultingStateProbability( const int index ){
	const int result_prob_size = get_result_probs_size();
	//Retrieve the original state-space size
	const int size = get_original_state_space_size();
	
	//Do conversion since originally it is double
	if( result_prob_size > 0 ){
		//Note that for a user state indexes start from 1
		if( 1 <= index && index <= size ){
			double * result = get_result_probs();
			print_state_prob(index-1, result_prob_size, result);
		}else{
			printf("ERROR: Invalid index %d, required to be in the [1, %d] interval.\n", index, size);
		}
	}else{
		printf("WARNING: There are NO results to print.\n");
	}
}

/**
* This method is used for printing state satisfyability result if any.
* I.e. if the state with the given index satisfyes the previously checked
* stateformula. In case get_result_bitset() == NULL or the index is out
* of bounds, the error messages are printed.
* @param index: the state index >= 1
*/
inline printResultingStateSatisfyability( const int index ){
	//Retrieve the original state-space size
	const int size = get_original_state_space_size();
	bitset * pBitset = get_result_bitset();

	//Do conversion since originally it is double
	if( pBitset != NULL && pBitset->n != 0 ){
		//Note that for a user state indexes start from 1
		if( 1 <= index && index <= size ){
			print_state(pBitset, index - 1);
		}else{
			printf("ERROR: Invalid index %d, required to be in the [1, %d] interval.\n", index, size);
		}
	}else{
		printf("WARNING: There are NO results to print.\n");
	}
}

/**
* This method creates and returns the bitset filled with ones.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @return the bitset of the length get_labeller()->ns that
*         contains all states, i.e. every bit in the set is assigned to 1.
*/
inline bitset * createTrueBitset(){
	//Starting the timer, if it has not been started yet
	startTimer();
	
	labelling *labellin = get_labeller();
	bitset *b = get_new_bitset(labellin->ns);
	fill_bitset_one( b );
	
	//The probability result, if any, is absolete now, so clean it.
	set_result_probs(0, NULL);
	
	return b;
}

/**
* This method creates and returns the bitset filled with zero.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @return the bitset of the length get_labeller()->ns that
*         contains no states, i.e. every bit in the set is assigned to 0.
*/
inline bitset * createFalseBitset(){
	//Starting the timer, if it has not been started yet
	startTimer();
	
	labelling *labellin = get_labeller();
	bitset *b = get_new_bitset(labellin->ns);
	
	//The probability result, if any, is absolete now, so clean it.
	set_result_probs(0, NULL);
	
	return b;
}

/**
* This method returns a set of states that satisfy atomic proposition 'label'.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @param label: the atomic proposition name.
* @return the set of states satisfying the atomic proposition 'label'.
*/
inline bitset * getStatesSetByLabel( const char * label ){
	//Starting the timer, if it has not been started yet
	startTimer();
	
	//If we just copy reference to the part of
	//labelling function (get_label_bitset) then
	//we will not be able to free bitset memory
	//later in the main loop of mcc.c
	labelling *labellin = get_labeller();
	bitset *b, *tmp_res;
	b = get_label_bitset(labellin, label);
	tmp_res = get_new_bitset(labellin->ns);
	if( b != NULL ){
		//Here we basically copy element of b in to a new bitset
		copy_bitset( b, tmp_res );
	}
	
	//The probability result, if any, is absolete now, so clean it.
	set_result_probs(0, NULL);

	//If there are no states with this label then
	//just return an empty set.
	return tmp_res;
}

/**
* This function is used for processing the formula
* connectives like AND, OR, IMPLIES. We have it separated
* because it is used in too many places here.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* NOTE: The input bitsets 'one' and 'two' are freed here.
* @param one: the first argument
* @param connective: the type of connective
* @param two: the second argument
* @result: the resulting bitset for 'one connective two'
*/
inline bitset * doConnective(bitset * one, int connective, bitset * two){
	bitset * pResultBitset = NULL;
	switch(connective)
	{
		case OR: pResultBitset = or(one, two); break;
		case AND: pResultBitset = and(one, two);break;
		case IMPLIES: pResultBitset = or(not_result(one), two);
	};
	free_bitset(one); free_bitset(two);
	
	//The probability result, if any, is absolete now, so clean it.
	set_result_probs(0, NULL);
	
	return pResultBitset;
}

/**
* This method is used to compute the complement of a bitset.
* NOTE: This method resets the probability result no NULL, using set_result_probs(...);
* @param pBitset: the input set
* @return the complement of the pBitset
*/
inline bitset * getComplementBitset(bitset * pBitset){
	bitset * pResultBitset = not_result(pBitset);
	
	//The probability result, if any, is absolete now, so clean it.
	set_result_probs(0, NULL);
	return pResultBitset;
}

/**
* This is a local method just for sorting the probab. results and
* creation of the satisfyability set.
* @param pResultBitSet: a preallocated (presumably empty) set that will be
*        filled with the states 'i' that satisfy 'result[i] comparator probBound'
* @param result: the array of probabilities
* @param probBound: the probability bound
* @param comparator: the comparator
* @param error_bound: the error bound that should be takein into account
*        while comparint the probabilities to the desired bound.
* WARNING: The Qureshi-Sanders Uniformization for Time- and Reward- bounded Until of CSRL
*          Provides us with the entire array of errors for each state, the same should be
*          happening with the Steady-state (Long-run) operator of PCTL, CSL, PRCTL, CSRL.
*          So having just one variable for error_bound here is not enough and this should
*          be fixed some time, Unfortunately we did not care about the error bounds much
*          before, so we have to implement a proper error-bound computation first.
*          Especially for the case of Nested formulas, where the error-bound derivation
*          formulas are just unknown.
*/
static inline sortOutStatesAccordingToProbs( bitset * pResultBitSet, double *result,  double probBound, int comparator, double error_bound){
	int i, size = get_labeller()->ns;
	//TODO: Since probabilistic results are computed with a certain error bound for 
	//the untill and steady-state (long-run) operators of PRCTL, PCTL, CSL and CSRL,
	//we have to check +- error bound bound:
	//1. Unbounded Untill of PRCTL, PCTL, CSL and CSRL.
	//2. Time-interval and Time-bounded Untill of CSL and CSRL.
	//3. Time and reward- bounded until of CSRL. Here though we should find
	//   out what is the error bound for Qureshi-Sanders uniformization and
	//   Tijms-Veldman discretization.
	//4. Long-run operator of PCTL, CSRL and Steady-state of CSL, CSRL.
	
	//Recompute the bounds based on the error_bound
	if( error_bound != 0 ){
		switch(comparator){
			case GREATER:
			case GREATER_OR_EQUAL:
				probBound -= error_bound;
				break;
			case LESS:
			case LESS_OR_EQUAL:
				probBound += error_bound;
		}
	}
	
	//Check for the bounds.
	for( i=0; i < size; i++ ){
		switch(comparator){
			case GREATER: if ( result[i] > probBound ) set_bit_val( pResultBitSet, i, BIT_ON );
				break;
			case GREATER_OR_EQUAL: if ( result[i] >= probBound ) set_bit_val( pResultBitSet, i, BIT_ON );
				break;
			case LESS: if ( result[i] <= probBound ) set_bit_val( pResultBitSet, i, BIT_ON );
				break;
			case LESS_OR_EQUAL: if ( result[i] < probBound ) set_bit_val( pResultBitSet, i, BIT_ON );
		}
	}
}

/**
* This function call model checking procedure for the steady state operator and then
* computes the set of states that satisfy the entire formula.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* @param pBitsetStateFormula the state formula result
* @param probBound the probability bound for the S operator
* @param comparator the comparator id from the parser
* @return the set of states satysfying the S formula
*/
inline bitset * getSteadyStateOperatorResultBitset(bitset * pBitsetStateFormula, double probBound, int comparator){
	const int size = get_labeller()->ns;
	bitset *b = get_new_bitset( size );
	double *result = NULL;
	
	if( isRunMode(PCTL_MODE) || isRunMode(PRCTL_MODE) || isRunMode(CSL_MODE) || isRunMode(CSRL_MODE) ){
		
		result = steady( pBitsetStateFormula );
		
		free_bitset( pBitsetStateFormula );
		
		//TODO: Since probabilistic results are computed with a certain error bound for 
		//the steady-state (long-run) operators of PRCTL, PCTL, CSL and CSRL,
		//we have to check +- error bound bound:
		//WARNING: Here the provided error bound is not quite correct,
		//it is tighter than it should be but at least it is something!
		//	S{<=>p}[phi], err = get_error_bound()
		//1. In case of ergodic MC we have to take:
		//	error_bound = err * | phi |
		//2. In case of Non ergodic MC it should be:
		//	error_bound = \Sum_j (y_j * err + x_ij * err * | phi_j | + err^2 * | phi_j |)
		//Where \Sum_j ( (x_ij +- err)*(y_j +- err * | phi_j |) )
		//is the way the probabiulity is computed for the state i.
		//	x_ij  - the probability to reach BSCC_j from state i
		//	y_j   - the the total steady state probability of phi states in BSCC_j
		//	phi_j - the phi states in BSCC_j
		double error_bound = get_error_bound();
		//Fill in the bitset b
		sortOutStatesAccordingToProbs( b, result, probBound, comparator, error_bound );
	} else {
		printf("ERROR: Steady-state formula S (long-run formula L) is only valid in CSL and CSRL (PCTL and PRCTL) mode.\n");
		result = (double *)calloc( size , sizeof(double));
	}
	
	//Store the probabilistic result since it can be needed for printing
	set_result_probs( size, result );

	return  b;
}

static BOOL isErrorBoundRequired = FALSE;

/**
* This function computes the set of states that satisfy the P formula.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* @param result the path formula result
* @param probBound the probability bound for the P operator
* @param comparator the comparator id from the parser
* @return the set of states satysfying the P formula
*/
inline bitset * getProbabilityOperatorResultBitset(double * result, double probBound, int comparator){
	const int size = get_labeller()->ns;
	bitset * b = get_new_bitset( size );
	
	//TODO: Since some probabilistic results are computed with a certain error
	//bound for the untill operators of PRCTL, PCTL, CSL and CSRL, we have
	//to check +- error bound bound:
	//1. Unbounded Untill of PRCTL, PCTL, CSL and CSRL.
	//2. Time-interval and Time-bounded Untill of CSL and CSRL.
	//3. Time and reward- bounded until of CSRL. Here though we should find
	//   out what is the error bound for Qureshi-Sanders uniformization and
	//   Tijms-Veldman discretization.
	//WARNING: Here the provided error bound is not quite correct,
	//it is tighter than it should be but at least it is something!
	//See comments for the
	//	sortOutStatesAccordingToProbs(...)
	//method as well.
	double error_bound = 0;
	if( isErrorBoundRequired ){
		error_bound = get_error_bound();
		isErrorBoundRequired = FALSE;
	}
	sortOutStatesAccordingToProbs( b, result, probBound, comparator, error_bound );
	
	//Store the probabilistic result since it can be needed for printing
	set_result_probs(size, result);
	return b;
}

/**
* This function calls model-checking algorithm for the unbounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @return the array of probabilities
*/
inline double * getUnboundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset){
	double * result = NULL;
	
	if( isRunMode(CSL_MODE) || isRunMode(PCTL_MODE) ){
		result = until(pPhiBitset, pPsiBitset, 0.0, 0.0, FALSE);
		isErrorBoundRequired = TRUE;
	}else if( isRunMode(PRCTL_MODE) || isRunMode(CSRL_MODE) ){
		result = until_rewards(pPhiBitset, pPsiBitset, 0.0, 0.0, 0.0, 0.0, FALSE);
		isErrorBoundRequired = TRUE;
	}else{
		printf("ERROR: Unbounded until formula U is valid only in PCTL, CSL, PRCTL and CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows , sizeof(double));
	}
	
	free_bitset(pPhiBitset); free_bitset(pPsiBitset);
	
	return result;
}

/**
* This function calls model-checking algorithm for the time-bounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @return the array of probabilities
*/
inline double * getTimeBoundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset, double t_bound_1, double t_bound_2){
	double * result = NULL;

	if( isRunMode(CSL_MODE) || isRunMode(PCTL_MODE) ){
		result = until(pPhiBitset, pPsiBitset, t_bound_1, t_bound_2, FALSE);
		isErrorBoundRequired = TRUE;
	}else if( isRunMode(PRCTL_MODE) || isRunMode(CSRL_MODE) ){
		result = until_rewards(pPhiBitset, pPsiBitset, t_bound_1, t_bound_2, 0.0, 0.0, FALSE);
		isErrorBoundRequired = TRUE;
	}else{
		printf("ERROR: Time-bounded until formula U is valid only in PCTL, CSL, PRCTL and CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows , sizeof(double));
	}
	
	free_bitset(pPhiBitset); free_bitset(pPsiBitset);
	
	return result;
}

/**
* This function calls model-checking algorithm for the time- and reward-bounded until
* NOTE: the pPhiBitset and pPsiBitset are freed here
* @param pPhiBitset the phi states bitset
* @param pPsiBitset the psi states bitset
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the array of probabilities
*/
inline double * getTimeAndRewardBoundedUntilProbability(bitset * pPhiBitset, bitset * pPsiBitset, double t_bound_1, double t_bound_2, double r_bound_1, double r_bound_2){
	double * result = NULL;
	
	if( isRunMode(PRCTL_MODE) || isRunMode(CSRL_MODE) ){
		result = until_rewards(pPhiBitset, pPsiBitset, t_bound_1, t_bound_2, r_bound_1, r_bound_2, FALSE);
	}else{
		printf("ERROR: Time- and reward-bounded until formula is valid only in PRCTL and CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows, sizeof(double) );
	}
	
	free_bitset(pPhiBitset); free_bitset(pPsiBitset);
	
	return result;
}

/**
* This function calls model-checking algorithm for the next formula
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X phi' is checked
* @return the array of probabilities
*/
inline double * getNextProbability(bitset * pBitset){
	double * result = NULL;
	
	if( isRunMode(PCTL_MODE) || isRunMode(CSL_MODE) ){
		result = next(pBitset, 0.0, 0.0);
	}else if( isRunMode(PRCTL_MODE) || isRunMode(CSRL_MODE) ){
		result = next_rewards(pBitset, 0.0, 0.0, 0.0, 0.0);
	}else{
		printf("ERROR: Next formula X is valid only in PCTL, CSL, PRCTL and CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows, sizeof(double) );
	}
	
	free_bitset(pBitset);
	
	return result;
}

/**
* This function calls model-checking algorithm for the time-bounded next formula
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @return the array of probabilities
*/
inline double * getTimeBoundedNextProbability(bitset * pBitset, double t_bound_1, double t_bound_2){
	double * result = NULL;
	
	if( isRunMode(CSL_MODE) ){
		result = next(pBitset, t_bound_1, t_bound_2);
	}else if( isRunMode(CSRL_MODE) ){
		result = next_rewards(pBitset, t_bound_1, t_bound_2, 0.0, 0.0);
	}else{
		printf("ERROR: Time-bounded next formula X is valid only in CSL and CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows, sizeof(double));
	}
	
	free_bitset(pBitset);
	
	return result;
}

/**
* This function calls model-checking algorithm for the time- and reward-bounded next formula
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param t_bound_1 lower time bound
* @param t_bound_2 upper time bound
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the array of probabilities
*/
inline double * getTimeAndRewardBoundedNextProbability(bitset * pBitset, double t_bound_1, double t_bound_2, double r_bound_1, double r_bound_2){
	double * result = NULL;
	
	if( isRunMode(CSRL_MODE) ){
		result = next_rewards(pBitset, t_bound_1, t_bound_2, r_bound_1, r_bound_2);
	}else{
		printf("ERROR: Time- and reward-bounded next formula X is valid only in CSRL mode.\n");
		result = (double *)calloc( get_state_space()->rows, sizeof(double));
	}
	
	free_bitset(pBitset);
	
	return result;
}

//Constants required for the getEECY(...) method and others that call it
#define EXPECTED_REWARD_RATE 0
#define INSTANTANEOUS_REWARD 1
#define EXPECTED_ACCUMULATED_REWARD 2

/**
* Universal method for E,C and Y formula
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param type the type which defines what formula it is: 
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
static inline bitset * getEECY(const int TYPE, bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2){
	const int size = get_labeller()->ns;
	int i;
	bitset *b = get_new_bitset(size);
	double *result = NULL;
	
	if( isRunMode(PRCTL_MODE) )
	{
		switch(TYPE){
			case EXPECTED_REWARD_RATE:
				result = ef(epoch, pBitset);
				break;
			case INSTANTANEOUS_REWARD:
				result = cf(epoch, pBitset);
				break;
			case EXPECTED_ACCUMULATED_REWARD:
				result = yf(epoch, pBitset);
				break;
			default:
				printf("ERROR: Undefined EECY type. This should not be happening!\n");
		}
		
		free_bitset(pBitset);

		for(i=0;i<size;i++){
			if( result[i] >= r_bound_1 && result[i] <= r_bound_2){
				set_bit_val(b, i, BIT_ON);
			}
		}
	}else{
		printf("ERROR: The expected reward rate formula E, instantaneous reward formula C and expected accumulated reward formula Y are valid only in PRCTL mode.\n");
	}
	
	//Store the probabilistic result since it can be needed for printing
	set_result_probs( (result == NULL ? 0 : size ), result);
	
	return b;
}

/**
* Calls model checking of the expected reward rate formula E.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
inline bitset * getExpectedRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2){
	return getEECY(EXPECTED_REWARD_RATE, pBitset, epoch, r_bound_1, r_bound_2);
}

/**
* Calls model checking of the instantaneous reward formula C.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
inline bitset * getInstantaneousRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2){
	return getEECY(INSTANTANEOUS_REWARD, pBitset, epoch, r_bound_1, r_bound_2);
}

/**
* Calls model checking of the expected accumulated reward Y.
* NOTE: It stores the probability result, if any, in runtime.c using set_result_probs(...)
* NOTE: the pBitset is freed here
* @param pBitset the bitset of states phi if 'X[t_bound_1,t_bound_2] phi' is checked
* @param epoch the epoch, for the long run expected reward rate should be set to 0 (ZERO)
* @param r_bound_1 lower reward bound
* @param r_bound_2 upper reward bound
* @return the bitset of states that satisfy the formula
*/
inline bitset * getExpectedAccumulatedRewardEpochRateInterval(bitset * pBitset, int epoch,  double r_bound_1, double r_bound_2){
	return getEECY(EXPECTED_ACCUMULATED_REWARD, pBitset, epoch, r_bound_1, r_bound_2);
}

//They should not be used anywhere else
#undef EXPECTED_REWARD_RATE
#undef INSTANTANEOUS_REWARD
#undef EXPECTED_ACCUMULATED_REWARD

/**
* This method should be used in case where cleaning of all previously
* gathered results is necessary.
*/
inline void cleanResults(){
	set_result_bitset(NULL);
	set_result_probs(0, NULL);
}
